/***
 * code is modified from the original code found at
 * https://github.com/Xinyuan-LilyGO/LilyGO-T-SIM7000G/blob/master/examples/Arduino_TinyGSM/AllFunctions/AllFunctions.ino
 */

#define TINY_GSM_MODEM_SIM7000

#include "ClientDataPackage.hpp"
#include "ConnectionManager.h"
#include "MessageType.h"
#include "QueueStruct.h"
#include "SDCardLogger.h"
#include "SPI.h"
#include "Utilities.h"
#include <ArduinoJson.h>
#include <Definitions.h>
#include <ESP32Time.h>
#include <NTPManager.h>
#include <TinyGsmClient.h>
#include <WiFi.h>
#include <esp_log.h>
#include <esp_now.h>
#include <queue>
#include <soc/rtc_cntl_reg.h>
#include <sys/unistd.h>

ESP32Time rtc;

#define TINY_GSM_RX_BUFFER 1024 // Set RX buffer to 1Kb
#define SerialAT Serial1

#ifdef DUMP_AT_COMMANDS // if enabled it requires the streamDebugger library
#include <StreamDebugger.h>
StreamDebugger debugger(SerialAT, Serial);
TinyGsm modem(debugger);
#else
TinyGsm modem(SerialAT);
#endif

const GPRSCredentials gprsCredentials = GPRSCredentials("m2m.public.at", "", "");

ConnectionManager connectionManager{modem, gprsCredentials,
                                    ConnectionManager::ModemPins{PIN_DTR, PIN_TX, PIN_RX, PWR_PIN}};

NTPManager ntpManager{modem};

uint8_t BROADCAST_MAC[6] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
esp_now_peer_info_t broadcast = {};
response announce = {};

SemaphoreHandle_t xMutex;

TaskHandle_t ESPNOWTask;

void handleDataReceive(const uint8_t *mac, const uint8_t *incomingData);
static std::queue<QueueStruct> queue;

void on_data_sent(const uint8_t *mac_addr, esp_now_send_status_t status) {
    // go to sleep
}

/**
 * @brief ESPNOW callback function that is called when data is received
 * @param mac
 * @param incomingData
 * @param len length of the incoming data in bytes
 */
void on_data_recv(const uint8_t *mac, const uint8_t *incomingData, int len) {
    ESPUtilities::checkPeerExistence(mac);

    ESPUtilities::sendResponse(mac);

    // requires queue and I am too lazy to extern declare the queue
    handleDataReceive(mac, incomingData);
}

[[noreturn]] void esp_loop() {
    while (true) {
    }
}

[[noreturn]] void ESPNOWReceiveTask(void *parameter) {
    esp_log_write(ESP_LOG_INFO, TAG_ESPNOW, "ESPNOWReceiveTask started on core %d\n", xPortGetCoreID());

    WiFi.mode(WIFI_STA);

    esp_log_write(ESP_LOG_DEBUG, TAG_ESPNOW, "Initialising ESPNow...\n");

    if (esp_now_init() != ESP_OK) {
        // initialization failed
        esp_log_write(ESP_LOG_ERROR, TAG_ESPNOW, "Initialising ESPNow FAILED\n");
        exit(ESP_FAIL);
    }

    esp_log_write(ESP_LOG_DEBUG, TAG_ESPNOW, "Initialising ESPNow SUCCESS\n");

    esp_now_register_recv_cb(on_data_recv);

    broadcast.channel = 0;
    broadcast.encrypt = false;
    memcpy(&broadcast.peer_addr, &BROADCAST_MAC, sizeof(BROADCAST_MAC));
    if (esp_now_add_peer(&broadcast) != ESP_OK) {
        esp_log_write(ESP_LOG_WARN, TAG_ESPNOW, "Failed to add Broadcast Host");
    }

    announce.type = hostChange;
    esp_read_mac(announce.mac, ESP_MAC_WIFI_STA);
    announce.time = rtc.getEpoch();
    if (esp_now_send(BROADCAST_MAC, (uint8_t *)&announce, sizeof(announce)) != ESP_OK) {
        esp_log_write(ESP_LOG_WARN, TAG_ESPNOW, "Failed to announce mac");
    } else {
        esp_log_write(ESP_LOG_DEBUG, TAG_ESPNOW, "Mac announced!");
    }

    esp_loop();
}

void setup() {
    // Set console baud rate
    Serial.begin(115200);
    delay(10);
    WRITE_PERI_REG(RTC_CNTL_BROWN_OUT_REG, 0); // disable detector
    try {
        SDUtilities::setupSDCard(SD_MISO, SD_MOSI, SD_SCLK, SD_CS);
    } catch (const SDSetupException &e) {
        SDUtilities::setSDAvailable(false);
        esp_log_write(ESP_LOG_ERROR, TAG_MAIN, "SD Card setup failed: %s\n", e.what());
    }

    SDCardLogger::printDebugToSerial(true);

    //	https://stackoverflow.com/questions/60442350/arduinos-esp-log-set-vprintf-does-not-work-on-esp32
    esp_log_set_vprintf(&SDCardLogger::vprintf_into_sd);
    esp_log_level_set("*", ESP_LOG_VERBOSE);
    esp_log_write(ESP_LOG_DEBUG, TAG_MAIN, "%s", WiFi.macAddress().c_str());

    turnOffLEDs();

    xMutex = xSemaphoreCreateMutex();

    delay(1000);

    // create ESPNOWReceiveTask. TODO: Until the UTC time is not synced, this will not add the correct time. If we
    // TODO: create the task after the time is synced, no messages will be received until synchronization is done
    xTaskCreatePinnedToCore(ESPNOWReceiveTask,   /* Function to implement the task */
                            "ESPNOWReceiveTask", /* Name of the task */
                            10000,               /* Stack size in words */
                            nullptr,             /* Task input parameter */
                            0,                   /* Priority of the task */
                            &ESPNOWTask,         /* Task handle. */
                            0);                  /* Core where the task should run */

    SerialAT.begin(UART_BAUD, SERIAL_8N1, PIN_RX, PIN_TX);

    // Restart takes quite some time
    // To skip it, call init() instead of restart()
}

bool isQueueEmpty() {
    bool emptyQueue;
    xSemaphoreTake(xMutex, portMAX_DELAY);
    emptyQueue = queue.empty();
    xSemaphoreGive(xMutex);
    return emptyQueue;
}

void loop() {

    // exit after 10 retries
    int earlyExitCounter = 0;

    // do while loop that takes one element of the queue each until the queue is empty
    while (!isQueueEmpty()) {
        xSemaphoreTake(xMutex, portMAX_DELAY);
        auto data = queue.front();
        queue.pop();
        xSemaphoreGive(xMutex);

        // get mac address as string
        String macAddress = ESPUtilities::getMacAddressAsString(data.mac);

        for (const auto &dataPackage : data.data) {
            // ignore padding messages
            if (dataPackage.getMeasurementData().getErrorType() == ErrorType::NULL_MESSAGE) {
                continue;
            }

            auto serverReadableString = dataPackageToServerReadableString(dataPackage, macAddress);

            try {
                SDUtilities::saveStringToSDCard(serverReadableString.c_str());
            } catch (const std::exception &e) {
                esp_log_write(ESP_LOG_ERROR, TAG_ESPNOW, "Failed to save data to SD card: %s", e.what());
            }

            // TODO: Try send or add to queue
        }

        if (earlyExitCounter > 10) {
            break;
        }
        earlyExitCounter++;
    }

    connectionManager.modemPowerOn();
    delay(5000L);

    // Restart takes quite some time
    // To skip it, call init() instead of restart()
    connectionManager.restartModem();

    announce.time = rtc.getEpoch();
    if (esp_now_send(BROADCAST_MAC, (uint8_t *)&announce, sizeof(announce)) != ESP_OK) {
        esp_log_write(ESP_LOG_WARN, TAG_ESPNOW, "Failed to announce mac\n");
    } else {
        esp_log_write(ESP_LOG_DEBUG, TAG_ESPNOW, "Mac announced!\n");
    }

    connectionManager.disableGPS();

    try {
        // may fail if already set
        connectionManager.setNetworkMode(LTE_ONLY);
        connectionManager.setPreferredMode(CAT_M);
    } catch (const std::exception &e) {
        esp_log_write(ESP_LOG_ERROR, TAG_GSM, "Error setting network mode: %s\n", e.what());
    }

    esp_log_write(ESP_LOG_DEBUG, TAG_GSM, "Modem Name: %s\n", modem.getModemName().c_str());

    esp_log_write(ESP_LOG_DEBUG, TAG_GSM, "Modem Info: %s\n", modem.getModemInfo().c_str());

    connectionManager.unlockSimCard();

    if (!connectionManager.waitForNetwork()) {
        esp_log_write(ESP_LOG_ERROR, TAG_GSM, "Network not available\n");
        return;
    }

    if (connectionManager.isNetworkConnected()) {
        esp_log_write(ESP_LOG_DEBUG, TAG_GSM, "Network connected\n");
    }

    delay(200);

    esp_log_write(ESP_LOG_DEBUG, TAG_GSM, "Connecting to %s\n", gprsCredentials.apn.c_str());
    if (!connectionManager.gprsConnect()) {
        esp_log_write(ESP_LOG_ERROR, TAG_GSM, "GPRS not connected\n");
        return;
    }

    connectionManager.logModemInformation();

    ntpManager.syncNTP("pool.ntp.org", "at.pool.ntp.org");

    try {
        ntpManager.writeModemTimeToRTC();
    } catch (const std::exception &e) {
        esp_log_write(ESP_LOG_ERROR, TAG_GSM, "Error writing time to rtc: %s\n", e.what());
    }

    //    RequestInformation requestInformation("GET", "vsh.pp.ua", "/TinyGSM/logo.txt", "");
    //    auto s = connectionManager.connect(80, requestInformation);
    //    esp_log_write(ESP_LOG_DEBUG, TAG_GSM, "Response: %s\n", s.c_str());

    connectionManager.gprsDisconnect();
    delay(5000L);

    if (!modem.isGprsConnected()) {
        esp_log_write(ESP_LOG_DEBUG, TAG_GSM, "GPRS disconnected\n");
    } else {
        esp_log_write(ESP_LOG_ERROR, TAG_GSM, "GPRS not disconnected\n");
    }

    connectionManager.modemPowerOff();

    delay(10000);
}
void handleDataReceive(const uint8_t *mac, const uint8_t *incomingData) {
    std::array<ClientDataPackage, 6> compressedDataPackage{};
    memcpy(&compressedDataPackage, incomingData, sizeof(ClientDataPackage) * 6);
    xSemaphoreTake(xMutex, portMAX_DELAY);
    queue.emplace(QueueStruct{compressedDataPackage, {mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]}});
    xSemaphoreGive(xMutex);
}